# coding: utf8

from django.contrib.auth import get_user_model
create_user = get_user_model().objects.create_user

from testutils import BaseTestCase

from .models import Account, Transaction


class AccountantBaseTestCase(BaseTestCase):
    email1 = 'test1@example.com'
    password1 = 'test'

    email2 = 'test2@example.com'
    password2 = 'test'

    def setUp(self):
        self.user1 = create_user(self.email1, self.password1)
        Account.objects.create(user=self.user1)
        self.user1_source = Account.GetPrimarySourceAccount(user=self.user1)
        self.user1_destination = Account.GetPrimaryDestinationAccount(
            user=self.user1)

        self.user2 = create_user(self.email2, self.password2)
        Account.objects.create(user=self.user2)
        self.user2_source = Account.GetPrimarySourceAccount(user=self.user2)
        self.user2_destination = Account.GetPrimaryDestinationAccount(
            user=self.user2)


class TestAccount(AccountantBaseTestCase):

    def test_new_user_balance(self):
        user = create_user('hope@dundermifflin.com', 'test')
        Account.objects.create(user=user)

        self.assertEqual(
            Account.GetPrimaryDestinationAccount(user).balance, 0.0)

        self.assertEqual(Account.GetPrimarySourceAccount(user).balance, 0.0)

    def test_balance(self):
        Transaction.objects.create(
            amount=50.0,
            source_account=self.user1_source,
            destination_account=self.user1_destination,
            is_settled=True,
        )

        self.assertEqual(self.user1_source.balance, 50.0)

    def test_transfer(self):
        Transaction.objects.create(
            amount=50.0,
            source_account=self.user1_source,
            destination_account=self.user1_destination,
            is_settled=True,
        )

        self.user1_source.transfer(50.0, self.user2_destination)
        self.assertEqual(self.user2_destination.balance, 50.0)

    def test_transfer_insufficient_balance(self):
        self.assertEqual(self.user1_source.balance, 0.0)

        with self.assertRaises(Account.InsufficientBalance):
            self.user1_source.transfer(50.0, self.user2_destination)


class TestBalanceClaim(AccountantBaseTestCase):

    def setUp(self):
        super(TestBalanceClaim, self).setUp()

        Transaction.objects.create(
            amount=50.0,
            source_account=self.user1_source,
            destination_account=self.user1_destination,
            is_settled=True,
        )

    def test_place_hold(self):
        self.assertEqual(self.user1_source.available_balance, 50.0)
        self.user1_source.place_hold(50.0)
        self.assertEqual(self.user1_source.available_balance, 0.0)

    def test_release_hold(self):
        self.assertEqual(self.user1_source.available_balance, 50.0)
        self.user1_source.place_hold(50.0)
        self.assertEqual(self.user1_source.available_balance, 0.0)
        self.user1_source.release_hold(50.0)
        self.assertEqual(self.user1_source.available_balance, 50.0)

    def test_transfer_with_no_available_funds(self):
        self.assertEqual(self.user1_source.available_balance, 50.0)
        self.user1_source.place_hold(50.0)
        self.assertEqual(self.user1_source.available_balance, 0.0)

        with self.assertRaises(Account.InsufficientBalance):
            self.user1_source.transfer(50.0, self.user2_destination)



# coding: utf8

from django.contrib import admin

from .models import Account, BalanceClaim, Transaction


class AccountAdmin(admin.ModelAdmin):
    list_display = (
        'user',
        'currency',
        'comment',
        'balance',
        'is_primary_destination',
        'is_primary_source',
        'created',
        'modified',
    )

admin.site.register(Account, AccountAdmin)


class BalanceClaimAdmin(admin.ModelAdmin):
    list_display = (
        'account',
        'amount',
        'is_released',
        'comment',
        'created',
        'modified',
    )

admin.site.register(BalanceClaim, BalanceClaimAdmin)


class TransactionAdmin(admin.ModelAdmin):
    list_display = (
        'source_account',
        'destination_account',
        'amount',
        'currency',
        'is_settled',
        'comment',
        'created',
        'modified',
    )

admin.site.register(Transaction, TransactionAdmin)



# coding: utf8

from distutils.core import setup
import os
import re

project_dir = 'accountant'


def parse_requirements(file_name):
    requirements = []
    for line in open(file_name, 'r').read().split('\n'):
        if re.match(r'(\s*#)|(\s*$)', line):
            continue
        if re.match(r'\s*-e\s+', line):
            requirements.append(re.sub(r'\s*-e\s+.*#egg=(.*)$', r'\1', line))
        elif re.match(r'\s*-f\s+', line):
            pass
        else:
            requirements.append(line)

    return requirements


def parse_dependency_links(file_name):
    dependency_links = []
    for line in open(file_name, 'r').read().split('\n'):
        if re.match(r'\s*-[ef]\s+', line):
            dependency_links.append(re.sub(r'\s*-[ef]\s+', '', line))

    return dependency_links

packages, data_files = [], []
root_dir = os.path.dirname(__file__)
if root_dir != '':
    os.chdir(root_dir)


def fullsplit(path, result=None):
    """
    Split a pathname into components (the opposite of os.path.join) in a
    platform-neutral way.
    """
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)


for dirpath, dirnames, filenames in os.walk(project_dir):
    # Ignore PEP 3147 cache dirs and those whose names start with '.'
    dirnames[:] = [
        d for d in dirnames if not d.startswith('.') and d != '__pycache__'
    ]
    if '__init__.py' in filenames:
        packages.append('.'.join(fullsplit(dirpath)))
    elif filenames:
        data_files.append([
            dirpath, [os.path.join(dirpath, f) for f in filenames]
        ])


setup(
    name='django-accountant',
    version='0.1.8',
    description='Django based account manager.',
    author='Ron Elliott',
    author_email='ronaldbelliott@gmail.com',
    install_requires=parse_requirements('requirements.txt'),
    dependency_links=parse_dependency_links('requirements.txt'),
    packages=packages,
    data_files=data_files,
)



# coding: utf8

import logging

from django.conf import settings
from django.db import models

from django_extensions.db.models import TimeStampedModel

log = logging.getLogger('accountant.models')


class Account(TimeStampedModel):
    """
    Represents an account that contains funds that may be transfered to other
    accounts.
    """

    comment = models.CharField(max_length=200, blank=True)
    currency = models.CharField(max_length=6, default='USD')
    is_primary_destination = models.BooleanField(default=True)
    is_primary_source = models.BooleanField(default=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='accounts')

    class InsufficientBalance(Exception):
        def __init__(self, balance, amount):
            self.balance = balance
            self.amount = amount

        def __str__(self):
            return str('%0.2f < %0.2f' % (self.balance, self.amount))

        def __unicode__(self):
            return '%0.2f < %0.2f' % (self.balance, self.amount)

    def __unicode__(self):
        if self.comment:
            return unicode('[%s]%s (%s)' % (
                self.currency, self.user, self.comment))
        return '[%s]%s' % (self.currency, self.user)

    @classmethod
    def GetMasterAccount(cls, currency='USD'):
        """ Returns the site master account. """

        return cls.objects.get(
            pk=settings.MASTER_ACCOUNT_PK, currency=currency)

    @classmethod
    def GetPrimaryDestinationAccount(cls, user, currency='USD'):
        """ Returns the given user's primary destination account. """

        try:
            return cls.objects.get(
                user=user, is_primary_destination=True, currency=currency
            )
        except cls.DoesNotExist:
            log.error(
                'Primary destination account for user: %s does not exist.'
                % user
            )

    @classmethod
    def GetPrimarySourceAccount(cls, user, currency='USD'):
        """ Returns the given user's primary source account. """

        try:
            return cls.objects.get(
                user=user, is_primary_source=True, currency=currency
            )
        except cls.DoesNotExist:
            log.error(
                'Primary source account for user: %s does not exist.'
                % user
            )

    def get_deposits(self, only_settled=True):
        """ Returns the deposit transactions. """

        return self.destination_transactions.filter(
            amount__gt=0, is_settled=only_settled)
    deposits = property(get_deposits)

    def get_withdrawals(self, only_settled=True):
        """ Returns the withdrawal transactions. """

        return self.source_transactions.filter(
            amount__lt=0, is_settled=only_settled)
    withdrawals = property(get_withdrawals)

    def get_refreshed_balance(self, only_settled=True):
        """
        Sums the source transactions less than zero and the destination
        transactions greater than zero.
        """

        negatives = sum(map(
            lambda t: t.amount,
            self.get_withdrawals(only_settled=only_settled)
        ))

        positives = sum(map(
            lambda t: t.amount,
            self.get_deposits(only_settled=only_settled)
        ))

        return float(positives + negatives)
    refreshed_balance = property(get_refreshed_balance)

    def get_balance(self, only_settled=True):
        """
        Sums the source transactions less than zero and the destination
        transactions greater than zero.
        """
        return self.get_refreshed_balance(only_settled=only_settled)
    balance = property(get_balance)

    def get_available_balance(self, only_settled=True):
        """
        Sums the source transactions less than zero and the destination
        transactions greater than zero.
        """
        balance = self.get_balance(only_settled=only_settled)
        holds = float(sum(map(
            lambda h: h.amount, self.claims.filter(is_released=False))))

        return balance - holds

    available_balance = property(get_available_balance)

    def place_hold(self, amount, comment=''):
        """ Creates a balance claim on this account. """

        return self.claims.create(amount=amount, comment=comment)

    def release_hold(self, amount, comment=''):
        """
        Creates a balance claim on this account, returning True if successful
        or False otherwise.
        """

        try:
            claim = self.claims.get(amount=amount, comment=comment)
            claim.release()
            return True
        except BalanceClaim.DoesNotExist:
            return False

    def transfer(self, amount, account, comment=''):
        """
        Transfers the amount from this account to the given account if the
        account has enough funds.
        """
        if self.available_balance < amount:
            log.error(
                'Account %d does not have enough funds to cover the transfer. '
                'Balance: %0.2f, Amount Requested: %0.2f, Comment: %s' % (
                    self.pk, self.available_balance, amount, comment))

            raise self.__class__.InsufficientBalance(
                self.available_balance, amount)

        Transaction.objects.create(
            amount=amount * -1,
            is_settled=True,
            source_account=self,
            destination_account=account,
            comment=comment,
        )

        return Transaction.objects.create(
            amount=amount,
            is_settled=True,
            source_account=self,
            destination_account=account,
            comment=comment,
        )


class BalanceClaim(TimeStampedModel):
    """ Represents a hold on a balance in an account. """

    account = models.ForeignKey('accountant.Account', related_name='claims')
    amount = models.DecimalField(decimal_places=8, max_digits=24)
    comment = models.CharField(max_length=200, blank=True)
    is_released = models.BooleanField(default=False, blank=True)

    def __unicode__(self):
        return '%f' % self.amount

    def release(self, commit=True):
        """ Releases the claim. """

        self.is_released = True

        if commit:
            self.save()

    def unrelease(self, commit=True):
        """ Unreleases the claim. """

        self.is_released = False

        if commit:
            self.save()


class Transaction(TimeStampedModel):
    """ Represents a transfer of funds between two accounts. """

    amount = models.DecimalField(decimal_places=8, max_digits=24)
    comment = models.CharField(max_length=200, blank=True, default='')
    is_settled = models.BooleanField(default=False)

    source_account = models.ForeignKey(
        'accountant.Account',
        related_name='source_transactions'
    )

    destination_account = models.ForeignKey(
        'accountant.Account',
        related_name='destination_transactions'
    )

    def __unicode__(self):
        return unicode('%f %s -> %s' % (
            self.amount,
            self.source_account,
            self.destination_account
        ))

    def get_currency(self):
        """ Returns the currency of the destination account. """

        return self.get_destination_currency()
    currency = property(get_currency)

    def get_destination_currency(self):
        """ Returns the currency of the destination account. """

        return self.destination_account.currency
    destination_currency = property(get_destination_currency)

    def get_source_currency(self):
        """ Returns the currency of the source account. """

        return self.destination_account.currency
    source_currency = property(get_source_currency)
